@@ -1,10 +1,11 @@
 # Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
-# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt
+# For details: https://github.com/coveragepy/coveragepy/blob/main/NOTICE.txt
 
 """Callback functions and support for sys.monitoring data collection."""
 
 from __future__ import annotations
 
+import collections
 import functools
 import inspect
 import os
@@ -14,18 +15,14 @@ import threading
 import traceback
 from dataclasses import dataclass
 from types import CodeType
-from typing import (
-    Any,
-    Callable,
-    NewType,
-    Optional,
-    cast,
-)
+from typing import Any, Callable, NewType, Optional, cast
 
 from coverage import env
 from coverage.bytecode import TBranchTrails, always_jumps, branch_trails
 from coverage.debug import short_filename, short_stack
+from coverage.exceptions import NoSource, NotPython
 from coverage.misc import isolate_module
+from coverage.parser import PythonParser
 from coverage.types import (
     AnyCallable,
     TFileDisposition,
@@ -39,6 +36,9 @@ from coverage.types import (
     TWarnFn,
 )
 
+# Only needed for some of the commented-out logging:
+# from coverage.debug import ppformat
+
 os = isolate_module(os)
 
 # pylint: disable=unused-argument
@@ -224,6 +224,10 @@ class SysMonitor(Tracer):
         # A list of code_objects, just to keep them alive so that id's are
         # useful as identity.
         self.code_objects: list[CodeType] = []
+
+        # Map filename:__name__ -> set(id(code_object))
+        self.filename_code_ids: dict[str, set[int]] = collections.defaultdict(set)
+
         self.sysmon_on = False
         self.lock = threading.Lock()
 
@@ -261,7 +265,7 @@ class SysMonitor(Tracer):
                         self.sysmon_branch_either,
                     )
                     register(
-                        events.BRANCH_LEFT,  # type:ignore[attr-defined]
+                        events.BRANCH_LEFT,
                         self.sysmon_branch_either,
                     )
             else:
@@ -282,6 +286,22 @@ class SysMonitor(Tracer):
             self.sysmon_on = False
             sys_monitoring.free_tool_id(self.myid)
 
+        if LOG:  # pragma: debugging
+            items = sorted(
+                self.filename_code_ids.items(),
+                key=lambda item: len(item[1]),
+                reverse=True,
+            )
+            code_objs = sum(len(code_ids) for _, code_ids in items)
+            dupes = code_objs - len(items)
+            if dupes:
+                log(f"==== Duplicate code objects: {dupes} duplicates, {code_objs} total")
+                for filename, code_ids in items:
+                    if len(code_ids) > 1:
+                        log(f"{len(code_ids):>5} objects: {filename}")
+            else:
+                log("==== Duplicate code objects: none")
+
     @panopticon()
     def post_fork(self) -> None:
         """The process has forked, clean up as needed."""
@@ -300,15 +320,17 @@ class SysMonitor(Tracer):
         return self.stats
 
     @panopticon("code", "@")
-    def sysmon_py_start(
-        self, code: CodeType, instruction_offset: TOffset
-    ) -> MonitorReturn:
+    def sysmon_py_start(self, code: CodeType, instruction_offset: TOffset) -> MonitorReturn:
         """Handle sys.monitoring.events.PY_START events."""
-        # Entering a new frame.  Decide if we should trace in this file.
         self._activity = True
         if self.stats is not None:
             self.stats["starts"] += 1
 
+        if code.co_name == "__annotate__":
+            # Type annotation code objects don't execute, ignore them.
+            return DISABLE
+
+        # Entering a new frame.  Decide if we should trace in this file.
         code_info = self.code_infos.get(id(code))
         tracing_code: bool | None = None
         file_data: TTraceFileData | None = None
@@ -323,7 +345,7 @@ class SysMonitor(Tracer):
                 frame = inspect.currentframe()
                 if frame is not None:
                     frame = inspect.currentframe().f_back  # type: ignore[union-attr]
-                    if LOG:
+                    if LOG:  # pragma: debugging
                         # @panopticon adds a frame.
                         frame = frame.f_back  # type: ignore[union-attr]
                 disp = self.should_trace(filename, frame)  # type: ignore[arg-type]
@@ -367,10 +389,16 @@ class SysMonitor(Tracer):
                             assert env.PYBEHAVIOR.branch_right_left
                             local_events |= (
                                 events.BRANCH_RIGHT  # type:ignore[attr-defined]
-                                | events.BRANCH_LEFT  # type:ignore[attr-defined]
+                                | events.BRANCH_LEFT
                             )
                         sys_monitoring.set_local_events(self.myid, code, local_events)
 
+                        if LOG:  # pragma: debugging
+                            if code.co_filename not in {"<string>"}:
+                                self.filename_code_ids[f"{code.co_filename}:{code.co_name}"].add(
+                                    id(code)
+                                )
+
         return DISABLE
 
     @panopticon("code", "@", None)
@@ -386,7 +414,7 @@ class SysMonitor(Tracer):
         code_info = self.code_infos.get(id(code))
         # code_info is not None and code_info.file_data is not None, since we
         # wouldn't have enabled this event if they were.
-        last_line = code_info.byte_to_line[instruction_offset]  # type: ignore
+        last_line = code_info.byte_to_line.get(instruction_offset)  # type: ignore
         if last_line is not None:
             arc = (last_line, -code.co_firstlineno)
             code_info.file_data.add(arc)  # type: ignore
@@ -433,9 +461,10 @@ class SysMonitor(Tracer):
         if not code_info.branch_trails:
             if self.stats is not None:
                 self.stats["branch_trails"] += 1
-            code_info.branch_trails = branch_trails(code)
+            multiline_map = get_multiline_map(code.co_filename)
+            code_info.branch_trails = branch_trails(code, multiline_map=multiline_map)
             code_info.always_jumps = always_jumps(code)
-            # log(f"branch_trails for {code}:\n    {code_info.branch_trails}")
+            # log(f"branch_trails for {code}:\n{ppformat(code_info.branch_trails)}")
         added_arc = False
         dest_info = code_info.branch_trails.get(instruction_offset)
 
@@ -445,7 +474,7 @@ class SysMonitor(Tracer):
             destination_offset = dest
             dests.add(destination_offset)
 
-        # log(f"{dest_info = }")
+        # log(f"dest_info = {ppformat(dest_info)}")
         if dest_info is not None:
             for arc, offsets in dest_info.items():
                 if arc is None:
@@ -459,11 +488,30 @@ class SysMonitor(Tracer):
         if not added_arc:
             # This could be an exception jumping from line to line.
             assert code_info.byte_to_line is not None
-            l1 = code_info.byte_to_line[instruction_offset]
-            l2 = code_info.byte_to_line.get(destination_offset)
-            if l2 is not None and l1 != l2:
-                arc = (l1, l2)
-                code_info.file_data.add(arc)  # type: ignore
-                # log(f"adding unforeseen {arc=}")
+            l1 = code_info.byte_to_line.get(instruction_offset)
+            if l1 is not None:
+                l2 = code_info.byte_to_line.get(destination_offset)
+                if l2 is not None and l1 != l2:
+                    arc = (l1, l2)
+                    code_info.file_data.add(arc)  # type: ignore
+                    # log(f"adding unforeseen {arc=}")
 
         return DISABLE
+
+
+@functools.lru_cache(maxsize=20)
+def get_multiline_map(filename: str) -> dict[TLineNo, TLineNo]:
+    """Get a PythonParser for the given filename, cached."""
+    try:
+        parser = PythonParser(filename=filename)
+        parser.parse_source()
+    except NotPython:
+        # The file was not Python. This can happen when the code object refers
+        # to an original non-Python source file, like a Jinja template.
+        # In that case, just return an empty map, which might lead to slightly
+        # wrong branch coverage, but we don't have any better option.
+        return {}
+    except NoSource:
+        # This can happen if open() in python.py fails.
+        return {}
+    return parser.multiline_map
